#!/usr/bin/python
# -*- coding: ascii -*-
"""
Version handler class.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import datetime
import os
import re
import subprocess

from dataclasses import dataclass
from pathlib import Path


# -->> Definitions <<------------------
PATH_THIS = Path(os.path.abspath(os.path.dirname(__file__)))


# -->> API <<--------------------------
@dataclass
class Version:
    """
    Firmware version definition class.
    """
    UVER_BUILD = 99999          #: Unversioned build number

    major: int = 99             #: Version major
    minor: int = 0              #: Version minor
    build: int = UVER_BUILD     #: Version build

    def __str__(self) -> str:
        """
        Version as string: major.minor.build
        """
        return f'{self.major}.{self.minor}.{self.build}'

    def update_build(self, path: Path) -> None:
        """
        Updates the build number for the current repo.

        :param Path path: Path to repo.
        """
        # Unversioned: do nothing.
        _ = path

    def save(self, file: Path, store_build: bool = False) -> None:
        """
        Stores the version number to the provided file.
        The version number will be stored in the format 'major.minor.build'.

        :param Path file: Path were the version text file will be stored.
        :param bool store_build: If true, the build will be stored. False by default.
        """
        if not file.parent.exists():
            raise ValueError(f'The provided path is not reachable: {file}')
        with file.open(mode='w') as ver_file:
            build = self.build if (store_build and self.build is not None) else 'X'
            ver_file.write(f'{self.major}.{self.minor}.{build}')

    @classmethod
    def load(cls, file: Path) -> 'Version':
        """
        Loads the version number from the provided file.
        The version number needs to be in the format 'major.minor.build'.

        :param Path file: Path to the version file.
        """
        if not (file.is_file() and file.exists()):
            raise ValueError(f'The provided path is not reachable: {file}')
        with file.open(mode='r') as ver_file:
            tmp = ver_file.read().split('.')
            ver = cls(
                major=int(tmp[0]),
                minor=int(tmp[1]),
                build=int(tmp[2]) if (tmp[2].upper() != 'X') else Version.UVER_BUILD
                )
            return ver


@dataclass
class VersionGit(Version):
    """
    Git version specialization.
    """
    def update_build(self, path: Path = Path(os.getcwd())) -> None:
        """
        Updates the build number for the current repo.

        :param Path path: Path to repo.
        """
        cmd = f'cd {path} && git rev-parse --short HEAD'
        ver = subprocess.check_output(cmd, shell=True).decode()
        if 'not a git' in ver.lower():
            self.build = self.UVER_BUILD
        else:
            self.build = int(ver, 16)


@dataclass
class VersionSVN(Version):
    """
    SVN version specialization.
    """
    def update_build(self, path: Path = Path(os.getcwd())) -> None:
        """
        Updates the build number for the current repo.

        :param Path path: Path to repo.
        """
        cmd = f'svnversion {path}'
        ver = subprocess.check_output(cmd, shell=True).decode()
        if 'unversioned directory' in ver.lower():
            self.build =  self.UVER_BUILD
        else:
            tmp = re.findall(pattern=r'\d+', string=ver)
            self.build = int(tmp[-1])


def export_version_c(ver: Version, author: str, note: str, file: Path) -> None:
    """
    Exports the version to a C header file.

    :param Version ver: Version instance.
    :param str author:  Author to be declared on the header file.
    :param str note:    Note to be added on the file documentation.
    :param Path file:   Path to version C header file.
    """

    if not (file.parent.exists() and file.parent.is_dir()):
        raise ValueError(f'The provided path is not reachable: {file}')

    template = (PATH_THIS / 'version_template.h').open(mode='r').read()
    template = template.format(
        file=file.name, author=author, note=note, date=f'{datetime.datetime.now():%x %X}',
        major=ver.major, minor=ver.minor,
        build=f'0x{ver.build:X}',
        version=f'"{ver.major}.{ver.minor}.{ver.build}"'
        )
    file.open('w').write(template)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Binary generation utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

from pathlib import Path
from typing import List, Optional, Tuple

import intelhex


# -->> Definitions <<------------------
RECORD_BYTES = 32


# -->> API <<--------------------------
def bin_to_hex(src: Path, off: int, out: Optional[Path] = None) -> intelhex.IntelHex:
    """
    Convert a binary file into an HEX.

    :param Path src: Path to source BIN file.
    :param int  off: BIN file address offset.
    :param Path out: Optional. If set, saves the generated hex file in the given path.

    :return: HEX file
    :rtype: intelhex.IntelHex
    """
    # Generate HEX
    tmp = intelhex.IntelHex()
    tmp.loadbin(fobj=f'{src}', offset=off)

    # Save if required
    if isinstance(out, Path):
        tmp.write_hex_file(f=f'{out}', byte_count=RECORD_BYTES)
    return tmp


def merge_bin(src: List[Tuple[Path, int]], out: Optional[Path] = None) -> intelhex.IntelHex:
    """
    Merge a group of binary files into an HEX.

    :param list src: List of tuples that contain:
        - BIN file path.
        - BIN file address offset
    :param Path out: Optional. If set, saves the generated hex file in the given path.

    :return: Merged HEX file.
    :rtype: intelhex.IntelHex
    """
    # Merge all BIN files
    tmp = intelhex.IntelHex()
    for file, addr in src:
        this = bin_to_hex(src=file, off=addr)
        tmp.merge(other=this, overlap='replace')

    # Save if required
    if isinstance(out, Path):
        tmp.write_hex_file(f=f'{out}', byte_count=RECORD_BYTES)
    return tmp


def merge_hex(src: List[Path], out: Optional[Path] = None) -> intelhex.IntelHex:
    """
    Merge a group of HEX files.

    :param list src: List of paths to HEX files.
    :param Path out: Optional. If set, saves the generated hex file in the given path.

    :return: Merged HEX file.
    :rtype: intelhex.IntelHex
    """
    # Merge all HEX files
    tmp = intelhex.IntelHex()
    for file in src:
        this = intelhex.IntelHex(source=f'{file}')
        tmp.merge(other=this, overlap='replace')

    # Save if required
    if isinstance(out, Path):
        tmp.write_hex_file(f=f'{out}', byte_count=RECORD_BYTES)
    return tmp



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Timing utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import time


# -->> Definitions <<------------------


# -->> API <<--------------------------
def time_elapsed(start: float) -> float:
    """
    Computes the elapsed time since the given timestamp.

    :param float start: Start timestamp.

    :returns: Time elapsed.
    :rtype: float
    """
    return time.time() - start



#!/usr/bin/python
# -*- coding: ascii -*-
"""
SDK threading utilities.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

from abc import ABC, abstractmethod
from functools import wraps
from queue import Empty, Queue
from threading import RLock, Thread
from typing import Callable, List, TypeVar

from .logger import SDK_LOG


# -->> Definitions <<------------------
RT = TypeVar('RT')


# -->> API <<--------------------------
def sync(lock_name: str) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """
    Decorator used to wrap a class method with a lock defined as
    instance/class attribute.

    :param str lock_name: Name of the lock attribute.

    :returns: Locked function.
    :rtype: Callable[[Callable[..., RT]], Callable[..., RT]]
    """
    def wrapper(func: Callable[..., RT]) -> Callable[..., RT]:
        """
        Wraps the decorated function to apply the lock.
        """
        @wraps(func)
        def synchronizer(self, *args, **kwargs) -> RT:
            """
            Applies a lock over the given function.
            """
            lock = self.__getattribute__(lock_name)
            with lock:
                return func(self, *args, **kwargs)
        return synchronizer
    return wrapper


class _ThreadWorker(Thread):
    """
    Thread pool worker.
    This represents a single thread on the pool. The thread is set as daemon or not based on
    the pool configurations.
    """
    def __init__(self, name: str, queue: Queue, timeout: float) -> None:
        """
        Class initialization.

        :param str name:        Worker name.
        :param Queue queue:     Queue to get the tasks from.
        :param float timeout:   Timeout for waiting for a task.
        """
        super().__init__(name=name)
        SDK_LOG.debug(f"Creating worker thread {self.name}")
        self._active = True
        self._queue = queue
        self._timeout = timeout

    def stop(self) -> None:
        """
        Stops the thread.
        """
        self._active = False

    def run(self) -> None:
        """
        Runs the worker logic.

        - Blocks until retrieves a task from the tasks queue (locked).
        - Checks that the task is a callable and execute.

        """
        # Worker execution
        while self._active:
            try:
                task = self._queue.get(timeout=self._timeout)
                if isinstance(task, AbstractThreadTask):
                    try:
                        task.execute()
                    except Exception as ex:
                        SDK_LOG.info(f"Caught exception while running task:\n"
                                     f"> Task   : {task}\n"
                                     f"> Raised : {ex.__class__.__name__} {ex}")
                    finally:
                        self._queue.task_done()
            except Empty:
                pass

        # Worker termination
        SDK_LOG.debug(f"Terminating worker thread {self.name}")


class AbstractThreadTask(ABC):
    """
    Thread task abstraction.
    Use this class to define how to execute a task inside the ThreadPool.
    """
    @abstractmethod
    def execute(self) -> None:
        """
        Execution called by the ThreadWorkers on the ThreadPool implementation.
        """


class SimpleThreadTask(AbstractThreadTask):
    """
    Simple thread task.
    Accepts a function to be executed by a worker on the ThreadPool.
    """
    def __init__(self, task: Callable[..., None], *args, name: str = 'Unnamed', **kwargs) -> None:
        """
        Task configuration:

        :param Callable[..., None] task:    Task function.
        :param str name:                    Task name.
        :param args:                        Task arguments.
        :param kwargs:                      Task keyword arguments.
        """
        # Check input
        if not callable(task):
            raise ValueError("The task is not a valid function")

        # Store attributes
        self._name   = name
        self._task   = task
        self._args   = args
        self._kwargs = kwargs

    def __repr__(self) -> str:
        """
        Representation string.
        """
        return f"{self.__class__.__name__}(name={self._name}, task={self._task.__name__})"

    def execute(self) -> None:
        """
        Execution called by the ThreadWorkers on the ThreadPool implementation.
        """
        self._task(*self._args, **self._kwargs)


class ThreadPool:
    """
    Simple thread pool implementation.
    Use queues to coordinate tasks among a set of worker threads.
    """
    def __init__(self, size: int, name: str, timeout: float = 60, daemon: bool = True) -> None:
        """
        :param int size:        Size of the thread pool.
        :param str name:        Thread pool name. Used as prefix on the workers as <name>_<worker_id>.
        :param float timeout:   Polling time used by the workers while waiting on a get() request
                                on the task queue. This value mainly affects the time used by the
                                threads to terminate when terminate() is called.
        :param bool daemon:     Set to true if the threads should immediately terminate when the
                                main thread exists.
        """
        # Input checking
        if size < 1:
            raise ValueError("Thread pool should have at least one worker")
        if timeout <= 0:
            raise ValueError("Polling timeout should be greater than zero")

        # Set attributes
        self._size = size
        self._daemon = daemon
        self._prefix = name
        self._timeout = timeout
        self._rlock = RLock()
        self._queue = Queue()
        self._workers: List[_ThreadWorker] = []

        # Create workers
        self._create_workers()

    @property
    @sync(lock_name='_rlock')
    def size(self) -> int:
        """
        Size of the thread pool.
        """
        return self._size

    def enqueue(self, task: AbstractThreadTask) -> None:
        """
        Enqueue a task on the thread pool to be executed by the workers.

        :param AbstractThreadTask task: Task to added to the queue.
        """
        # Check input
        if not isinstance(task, AbstractThreadTask):
            raise ValueError(f"The task need to be {AbstractThreadTask.__name__} or a subclass of it")
        # Enqueue
        self._queue.put(task)

    @sync(lock_name='_rlock')
    def stop(self) -> None:
        """
        Wait for the task queue to be completed and stops all the workers.
        """
        SDK_LOG.debug("Waiting for all the tasks to be handled...")
        self._queue.join()
        SDK_LOG.debug("Terminating worker threads...")
        for worker in self._workers:
            worker.stop()

    @sync(lock_name='_rlock')
    def _create_workers(self) -> None:
        """
        Create all the pool workers.
        """
        while len(self._workers) < self._size:
            worker = _ThreadWorker(name=f"{self._prefix}_{(len(self._workers) + 1)}", queue=self._queue, timeout=self._timeout)
            worker.setDaemon(self._daemon)
            worker.start()
            self._workers.append(worker)


# -->> Instances <<--------------------
#: Embutils internal thread pool
SDK_TP = ThreadPool(size=4, name='EMBUTILS_Thread_', timeout=30)



#!/usr/bin/python
# -*- coding: ascii -*-
"""
COBS encoding/decoding implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""


# -->> Definitions <<------------------


# -->> API <<--------------------------
class COBS:
    """
    Consistent Overhead Byte Stuffing (COBS) encoding/decoding utilities.
    """
    class DecodeException(Exception):
        """
        COBS decoding exception.
        """

    @staticmethod
    def encode(data: bytearray = bytearray()) -> bytearray:
        """
        Encode a byte array using Consistent Overhead Byte Stuffing (COBS).

        .. note::
            * Encoding guarantees non-zero bytes on the output array.
            * An empty string is encoded to [0x01]

        :param bytearray data: Bytes to be encoded.

        :returns: Encoded byte array.
        :rtype: bytearray
        """
        # Verify that we have bytes to encode...
        if len(data) == 0:
            return bytearray([0x01])

        # Encode
        frame = bytearray()
        add_zero = False
        idx_start = 0
        idx_end = 0
        for byte in data:
            # Compute code
            code = idx_end - idx_start + 1

            # Process the current byte
            if byte == 0x00:
                # Add a zero: Block termination
                #   - If this is the last message: Add another zero to show that is ready!
                #   - We need to append < 0xFE bytes!
                #   - Append the code (bytes available).
                add_zero = True
                frame.append(code)
                frame += data[idx_start:idx_end]
                idx_start = idx_end + 1

            elif code == 0xFE:
                # Special case: Send the full block with non-zero bytes
                #   - We have 0xFE bytes available!
                #   - The block header is set to 0xFF
                #   - Don't need to add a zero termination!
                add_zero = False
                frame.append(0xFF)
                frame += data[idx_start:(idx_end + 1)]
                idx_start = idx_end + 1

            # Explore the next byte...
            idx_end += 1

        # Finish the packet:
        if (idx_end != idx_start) or add_zero:
            # If needed:
            #   - Add the block header code (if no bytes missing: 0x01)
            #   - Add the missing bytes
            code = idx_end - idx_start + 1
            frame.append(code)
            frame += data[idx_start:idx_end]

        return frame

    @staticmethod
    def decode(data: bytearray) -> bytearray:
        """
        Decodes a byte array that was encoded using Consistent Overhead Byte
        Stuffing (COBS).

        :param bytearray data: Bytes to be decoded.

        :returns: Decoded byte array.
        :rtype: bytearray

        :raises COBS.DecodeException: Encoded data is invalid.
        """
        # Verify that he have enough bytes
        if len(data) == 0:
            return bytearray()

        msg = bytearray()
        idx = 0

        while idx < len(data):
            # Get the frame block code
            code = data[idx]
            if code == 0x00:
                raise COBS.DecodeException('Zero byte found in input!')

            # Get the data
            idx += 1
            end = idx + code - 1

            tmp = data[idx:end]
            if 0x00 in tmp:
                raise COBS.DecodeException('Zero byte found in input!')

            # Append it to the message
            msg += tmp

            # Update and check index
            idx = end
            if idx > len(data):
                raise COBS.DecodeException('Not enough bytes to process!')

            if idx < len(data):
                # In range, add zero if needed
                if code < 0xFF:
                    msg.append(0x00)

            else:
                # All the bytes processed
                break

        return msg



#!/usr/bin/python
# -*- coding: ascii -*-
"""
SDK logger implementation.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

from collections import namedtuple
from logging import Logger as _Logger, Formatter, StreamHandler
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING

from .enum import IntEnum


# -->> Definitions <<------------------


# -->> API <<--------------------------
class Logger(_Logger):
    """
    Logger implementation wrapper.
    This class simplifies the logger formatting and level definition for the SDK.
    """
    class Level(IntEnum):
        """
        Available logger message levels.
        """
        CRITICAL = CRITICAL
        DEBUG = DEBUG
        ERROR = ERROR
        INFO = INFO
        WARNING = WARNING

    #: Logger entries format definition
    Format = namedtuple('Format', ['style', 'pattern'])

    #: Default entry format
    DEF_FMT = Format(style='{', pattern='{created:.05f}: {name:<8s}: {levelname:<8s}: {module:<20s}: {message:s}')

    def __init__(self, name: str = '', level: Level = Level.DEBUG, fmt: Format = DEF_FMT) -> None:
        """
        Logger configuration.
        Applies the log entries formatting and initial log level.

        :param str name:  Logger name.
        :param int level: Logger initial message name.
        :param Logger.Format fmt: Log entries formatting. By default :attr:`FMT_DEFAULT`.
        """
        # Initialize logger
        super().__init__(name=name, level=level)

        # Define formatter
        if fmt is None:
            fmt = self.DEF_FMT

        # Disable logger by default
        self.disable()

        # Configure format
        fmt  = Formatter(fmt=fmt.pattern, style=fmt.style)
        hdlr = StreamHandler()
        hdlr.setFormatter(fmt=fmt)
        self.addHandler(hdlr=hdlr)

    def enable(self) -> None:
        """
        Enable logger.
        """
        self.disabled = False

    def disable(self) -> None:
        """
        Disable logger.
        """
        self.disabled = True

    def set_level(self, level: Level = Level.DEBUG) -> None:
        """
        Set the log messages level.

        :param int level: Log level.
        """
        self.setLevel(level=level)


# -->> Instances <<--------------------
#: Embutils internal logger
SDK_LOG = Logger(name='EMBUTILS')



#!/usr/bin/python
# -*- coding: ascii -*-
"""
Serial interface implementation. This class represents the closer point to the
user. Here the developer needs to complete the interface with the specific
applications commands.

:date:      2021
:author:    Christian Wiche
:contact:   cwichel@gmail.com
:license:   The MIT License (MIT)
"""

import time
from typing import Protocol, Optional

from ..utils import SDK_LOG, AbstractSerialized, EventHook, time_elapsed
from .stream import Stream


# -->> Definitions <<------------------
class ResponseCheckCallback(Protocol):
    """
    Response check callback signature:
    AbstractSerialized -> bool
    """
    def __call__(self, item: AbstractSerialized) -> bool: ...


# -->> API <<--------------------------
class Interface:
    """
    Serial command interface implementation.
    This class should implement all the methods to interact with the target device.

    Available events:

    #.  **on_receive:** This event is emitted when an object is received and
        deserialized from the serial device. Subscribe using callbacks with
        syntax::

            def <callback>(item: AbstractSerialized)

    #.  **on_connect:** This event is emitted when the system is able to
        connect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_reconnect:** This event is emitted when the system is able to
        reconnect to the device. Subscribe using callbacks with syntax::

            def <callback>()

    #.  **on_disconnect:** This event is emitted when the system gets
        disconnected from the device. Subscribe using callbacks with syntax::

            def <callback>()

    """
    #: Interface command pull period
    PERIOD_PULL_S = 0.005

    #: Interface join period
    PERIOD_JOIN_S = 0.5

    #: Interface command response timeout
    TIMEOUT_RESPONSE_S = 0.5

    def __init__(self, stream: Stream) -> None:
        """
        Class initialization.

        :param Stream stream:   Stream used to run the interface
        """
        # Response timeout configuration
        self._timeout = self.TIMEOUT_RESPONSE_S

        # Public events
        self.on_receive     = EventHook()
        self.on_connect     = EventHook()
        self.on_reconnect   = EventHook()
        self.on_disconnect  = EventHook()

        # Initialize stream and attach callbacks
        self._stream = stream
        self._stream.on_connect     += self.on_connect.emit
        self._stream.on_reconnect   += self.on_reconnect.emit
        self._stream.on_disconnect  += self.on_disconnect.emit
        self._stream.on_receive    += self._process

        SDK_LOG.info(f'Interface initialized on: {self._stream.device}')

    @property
    def stream(self) -> Stream:
        """
        Stream handler.
        """
        return self._stream

    @property
    def timeout(self) -> float:
        """
        Message response timeout.
        """
        return self._timeout

    @timeout.setter
    def timeout(self, timeout: float) -> None:
        """
        Message response timeout setter.

        :param float timeout: Timeout in seconds.
        """
        if timeout <= 0.0:
            raise ValueError('The response timeout needs to be greater than zero.')
        self._timeout = timeout

    def _process(self, item: AbstractSerialized) -> None:
        """
        Method used to handle the received items for processing.

        :param AbstractSerialized item: item to be processed.
        """
        # Do special actions:

        # Pass item to user:
        self.on_receive.emit(item=item)

    def transmit(self,
                 send: AbstractSerialized,
                 logic: ResponseCheckCallback = None,
                 timeout: float = None) -> Optional[AbstractSerialized]:
        """
        Send a item and, if required, wait for a response that complies with
        the defined logic.

        The response detection logic should have the following syntax::

            def <callback>(AbstractSerialized) -> bool:

        Where the input is the item received from the device.

        :param AbstractSerialized send:                     Item to be sent.
        :param Callable[[AbstractSerialized], bool] logic:  Response logic. Defines if a response is detected.
        :param float timeout:                               Response timeout. By default the interface setting.

        :returns: None if timeout or no response is detected, response item otherwise.
        :rtype: Optional[AbstractSerialized]
        """
        recv    = None
        timeout = self._timeout if (timeout is None) else timeout

        # Receive logic callback
        def on_received(item: AbstractSerialized) -> None:
            nonlocal send, recv
            if logic(item):
                recv = item

        # Prepare response logic
        if logic:
            self.on_receive += on_received

        # Prepare and send
        SDK_LOG.debug('Sending item...')
        self._stream.send(item=send)

        # Define if response is needed
        if logic:
            # Wait for response
            SDK_LOG.debug('Waiting for response...')
            tm_start = time.time()
            while not recv and (time_elapsed(tm_start) < timeout):
                time.sleep(self.PERIOD_PULL_S)
            self.on_receive -= on_received

            # Check data
            state = "Received" if recv else "Timeout"
            SDK_LOG.debug(f'Item response: {state} after {time_elapsed(start=tm_start):.03f}[s]')
            return recv

        # Return nothing if we don't need response
        SDK_LOG.debug('Response not needed.')
        return None

    def stop(self) -> None:
        """
        Stops the stream process.
        """
        self._stream.stop()

    def join(self) -> None:
        """
        Maintains the process alive while the interface is working. The loop will
        exit if the interface serial process gets stopped.
        """
        while self._stream.is_alive:
            time.sleep(self.PERIOD_JOIN_S)



from .binary import bin_to_hex, merge_bin, merge_hex
from .bytes import bitmask, reverse_bits, reverse_bytes
from .cobs import COBS
from .crc import CRC
from .enum import IntEnum
from .events import EventHook
from .logger import Logger, SDK_LOG
from .math import closest_pow, closest_multi
from .serialized import AbstractSerialized, AbstractSerializedCodec
from .subprocess import execute
from .threading import AbstractThreadTask, SimpleThreadTask, ThreadPool, sync, SDK_TP
from .time import time_elapsed
